from typing import Any, Dict

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import pack_int

class ObjPlayerInfoRace(PrettyPrint):
    def __init__(
            self,
            race_start_tick: int = 0
    ) -> None:
        self.item_name: str = 'obj.player_info_race'
        self.type_id: int = 23
        self.id: int = 0
        self.size: int = 1

        self.race_start_tick: int = race_start_tick

    def __iter__(self):
        yield 'item_name', self.item_name
        yield 'type_id', self.type_id
        yield 'id', self.id
        yield 'size', self.size

        yield 'race_start_tick', self.race_start_tick

    def to_dict_payload_only(self) -> Dict[str, Any]:
        return {
            'race_start_tick': self.race_start_tick
        }

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, unpacker: Unpacker) -> bool:
        # TODO: it is weird to unpack the size here because
        #       it is not really part of the payload
        #       but this is one of the two edge cases messages
        #       this patch is also why it has to be excluded from the generation script
        self.size = unpacker.get_int()

        self.race_start_tick = unpacker.get_int()
        return True

    def pack(self) -> bytes:
        return pack_int(self.size) + \
            pack_int(self.race_start_tick)



# TODO: not super sure yet how many session structs we need
#       should there be one for 0.7 too? One for ddnet too?
#       or just version fields?
class Session7:
    def __init__(self):
        self.client_token = b'\xff\xff\xff\xff'
        self.server_token = b'\xff\xff\xff\xff'

        # The amount of vital chunks received
        self.ack = 0

        # The amount of vital chunks sent
        self.sequence = 0

        # The amount of vital chunks acknowledged by the peer
        self.peer_ack = 0



from twnet_parser.constants import NET_MAX_SEQUENCE

def seq_in_backroom(sequence: int, ack: int):
    """
    Only used for chunks where the sequence number does not match the expected value
    to decide whether to drop known chunks silently or request resend if something got lost
    The expected value would be the `(last_ack + 1) % DDNET_MAX_SEQUENCE`

    The argument `sequence` is the sequence number of the incoming chunk

    The argument `ack` is the expected sequence number

    true - if the sequence number is already known and the chunk should be dropped
    false - if the sequence number is off and we need to request a resend of lost chunks
    """
    bottom = ack - (NET_MAX_SEQUENCE / 2)
    if bottom < 0:
        if sequence <= ack:
            return True
        if sequence >= (bottom + NET_MAX_SEQUENCE):
            return True
    elif bottom <= sequence <= ack:
        return True
    return False



# return type Snapshot from within Snapshot class
# remove this after switching to python 3.14
from __future__ import annotations

import copy
from typing import Literal, Optional
from twnet_parser import snap_item
from twnet_parser.obj7 import GAME_DATA_RACE, PLAYER_INFO_RACE
from twnet_parser.packer import Unpacker, pack_int
from twnet_parser.snap_item import SnapItem
from twnet_parser.item_matcher.match_snap7 import match_item as match_item7

# TODO: this is horrible
def get_item_payload(item: SnapItem) -> list[int]:
    ints: list[int] = []
    unpacker = Unpacker(item.pack())
    for _ in range(0, item.size):
        ints.append(unpacker.get_int())
    return ints


# TODO: don't undiff items the slowest possible way
#
#	there should be way to do it like the C++ implementation
#	which does no unpacking or repacking before applying the diff
def undiff_item_slow(old_item: SnapItem, diff_item: SnapItem) -> SnapItem:
    if snap_item is None:
        return diff_item

    if old_item.type_id != diff_item.type_id:
        raise ValueError("Error: can not diff items of different type")
    if old_item.size != diff_item.size:
        raise ValueError("Error: can not diff items of different size")

    old_payload = get_item_payload(old_item)
    diff_payload = get_item_payload(diff_item)
    new_payload: bytes = b''

    for i in range(0, old_item.size):
        diff_applied = old_payload[i] + diff_payload[i]
        new_payload += pack_int(diff_applied)

    unpacker = Unpacker(new_payload)
    old_item.unpack(unpacker)
    return old_item

# TODO: move this??
def item_key(item: SnapItem) -> int:
    return (item.type_id << 16) | (item.id & 0xffff)

class Snapshot:
    """
    The snapshot is representing the current state
    of the most gameplay relevant moving things.

    It contains player, projectile and pickup positions.
    And a few more things. It does NOT include any map data.

    Snapshots are always deltas to a previous snapshot.
    The first snapshot is not a delta
    (or a delta to the empty snapshot however you want to see it).

    Snapshots work technically very similar in 0.7 and 0.6
    so this class covers both of them. If you want to understand more
    about snapshots you can read about it here:
    https://chillerdragon.github.io/teeworlds-protocol/07/snap_items.html
    """

    MAX_TYPE: int = 0x7fff
    MAX_ID: int = 0xffff
    MAX_PARTS: int = 64
    MAX_SIZE: int = MAX_PARTS * 1024
    MAX_PACK_SIZE: int = 900

    def __init__(
            self,
            version: Literal['0.6', '0.7'],
            num_removed_items: int = 0,
            num_item_deltas: int = 0
    ) -> None:
        self._version: Literal['0.6', '0.7'] = version
        self.num_removed_items: int = num_removed_items
        self.num_item_deltas: int = num_item_deltas
        self.removed_item_keys: list[int] = []
        self.items: list[SnapItem] = []
        self.crc = 0

    @property
    def version(self) -> Literal['0.6', '0.7']:
        return self._version

    def match_item(self, unpacker: Unpacker) -> Optional[SnapItem]:
        if self.version == '0.7':
            return match_item7(unpacker)
        raise ValueError(f'Error: {self.version} snapshots are not implemented yet')

    # expects the int compressed
    # data field of the snap message
    # not the whole snap message with crc
    # and the other fields
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        self.num_removed_items = unpacker.get_int()
        self.num_item_deltas = unpacker.get_int()
        unpacker.get_int() # unused by tw 0.7 NumTempItems

        self.removed_item_keys = []
        for _ in range(0, self.num_removed_items):
            self.removed_item_keys.append(unpacker.get_int())

        while unpacker.remaining_size() > 2:
            item = self.match_item(unpacker)
            if item is None:
                continue
            item.id = unpacker.get_int()
            item.unpack(unpacker)
            # print(f"got item size={item.size} type={item.type_id}")
            self.items.append(item)

        self.regenerate_crc()
        return True

    def pack(self) -> bytes:
        return b''

    # the key is one integer holding both type and id
    def get_item_at_key(self, key: int) -> Optional[SnapItem]:
        for item in self.items:
            if key == item_key(item):
                return item
        return None

    # the key is one integer holding both type and id
    def get_item_index(self, key: int) -> Optional[int]:
        for index, item in enumerate(self.items):
            if key == item_key(item):
                return index
        return None


    # TODO: this is wasting clock cycles for no reason
    #
    #	the crc is all snap item payload integers summed together
    #	it does not have to be perfectly optimized
    #	but repacking every item to get its payload summed is horrible
    #
    #	i also had another approach with reflect where every snap object would implement
    #	Crc() on them selfs
    #	but reflect is messy and especially the enum types got annoying to sum
    #	because they require specific casting
    def crc_item(self, item: SnapItem) -> int:
        # TODO: remove once crc is confirmed to be working
        # print(f"getting crc from item {item.item_name} that has a size of {item.size}")

        unpacker = Unpacker(item.pack())

        if self.version == '0.7':
            if item.type_id in (GAME_DATA_RACE, PLAYER_INFO_RACE):
                # the backwards compatibility size
                # is not part of the payload
                # and is not used to compute the crc
                unpacker.get_int()
        if item.item_name == 'obj.unknown':
            unpacker.get_int() # pop size

        # TODO: remove once crc is confirmed to be working
        # print(f" unpacker remaining data size {unpacker.remaining_size()}")

        crc = 0
        for _ in range(0, item.size):
            # TODO: remove once crc is confirmed to be working
            # print(f"   unpacker popped field {i} remaining data size {unpacker.remaining_size()}")

            # 32 bit overflow to match the reference implementation
            # not sure how correct this is should probably be double checked
            # i got it from here https://stackoverflow.com/a/16745422
            crc = (unpacker.get_int() + crc) & 0xffffffff
        return crc

    # from has to be the old snapshot we delta against
    # and the unpacker has to point to the payload of the new delta snapshot
    # the payload starts with NumRemovedItems
    #
    # it returns the new full snapshot with the delta applied to the from
    def unpack_delta(self, unpacker: Unpacker) -> Snapshot:
        # TODO: add all the error checking the C++ reference implementation has

        snap = Snapshot(self.version)
        snap.num_removed_items = unpacker.get_int()
        snap.num_item_deltas = unpacker.get_int()
        unpacker.get_int() # _zero

        deleted_keys: list[int] = []

        for _ in range(0, snap.num_removed_items):
            deleted_keys.append(unpacker.get_int())

        for copy_item in self.items:
            keep = True

            for deleted_key in deleted_keys:
                if deleted_key == item_key(copy_item):
                    keep = False
                    break

            if keep:
                snap.items.append(copy.deepcopy(copy_item))

        for _ in range(0, snap.num_item_deltas):
            item: Optional[SnapItem] = copy.deepcopy(self.match_item(unpacker))
            if item is None:
                raise ValueError("Error: failed to unpack snap item")

            item.id = unpacker.get_int()
            # print(f"unpacking item {item.item_name} typeid={item.type_id} i={i} size={item.size}")
            item.unpack(unpacker)

            key = (item.type_id << 16) | (item.id & 0xffff)
            old_item = copy.deepcopy(self.get_item_at_key(key))

            if not old_item:
                snap.items.append(item)
            else:
                item = undiff_item_slow(old_item, item)
                idx = snap.get_item_index(key)
                if idx is None:
                    raise ValueError( \
                            "Error: snap item update failed. " +
                            f"Item with key={key} type_id={item.type_id} id={item.id} not found."
                    )
                snap.items[idx] = item

        if unpacker.remaining_size() > 0:
            raise ValueError("Error:" \
                f"unexpected remaining size {unpacker.remaining_size()} after snapshot unpack")

        snap.regenerate_crc()
        return snap

    def regenerate_crc(self):
        crc = 0
        for item in self.items:
            # 32 bit overflow to match the reference implementation
            # not sure how correct this is should probably be double checked
            # i got it from here https://stackoverflow.com/a/16745422
            crc = (self.crc_item(item) + crc) & 0xffffffff
        self.crc = crc
        return crc




from typing import Any, Dict

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import pack_int

# special meta item that represents unknown items
class ObjUnknown(PrettyPrint):
    def __init__(self) -> None:
        self.item_name: str = 'obj.unknown'
        self.type_id: int = 0
        self.id: int = 0
        self.size: int = 0

        self.fields: list[int] = []

    def __iter__(self):
        yield 'item_name', self.item_name
        yield 'type_id', self.type_id
        yield 'id', self.id
        yield 'size', self.size

        yield 'fields', self.fields

    def to_dict_payload_only(self) -> Dict[str, Any]:
        return {
            'fields': self.fields
        }

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, unpacker: Unpacker) -> bool:
        # TODO: this is a bit weird
        #       the size is kinda part of the "header"
        #       but only for unknown items
        #       so the unknown item uses the same interface
        #       which promises to unpack the "payload" only
        #       is the size now header or payload?
        self.size = unpacker.get_int()

        self.fields = []
        for _ in range(0, self.size):
            self.fields.append(unpacker.get_int())
        return True

    def pack(self) -> bytes:
        data = pack_int(self.size)
        for i in range(0, self.size):
            data += pack_int(self.fields[i])
        return data



#!/usr/bin/env python

# pylint: disable=duplicate-code

from typing import Union, cast, Optional, Literal, Iterator, Any
import json

from twnet_parser.packer import pack_int
from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.net_message import NetMessage
from twnet_parser.ctrl_message import CtrlMessage
from twnet_parser.connless_message import ConnlessMessage
from twnet_parser.constants import NET_MAX_SEQUENCE, NET_PACKETVERSION

import twnet_parser.huffman
import twnet_parser.serialize

PACKETFLAG6_UNUSED = 1
PACKETFLAG6_TOKEN = 2
PACKETFLAG6_CONTROL = 4
PACKETFLAG6_CONNLESS = 8
PACKETFLAG6_RESEND = 16
PACKETFLAG6_COMPRESSION = 32

# not sure if this is correct
# i get big 3 byte vibes from 0.6
# and then there is the 0.6.5 token extension
CONNLESS_PACKET_HEADER6_SIZE = 6

class PacketFlags6(PrettyPrint):
    def __init__(self) -> None:
        self.token: Optional[bool] = None
        self.control: Optional[bool] = None
        self.resend: Optional[bool] = None
        self.compression: Optional[bool] = None
        self.connless: Optional[bool] = None

    def __iter__(self):
        flags = []
        if self.token:
            flags.append('token')
        if self.control:
            flags.append('control')
        if self.resend:
            flags.append('resend')
        if self.compression:
            flags.append('compression')
        if self.connless:
            flags.append('connless')
        return iter(flags)

    def __repr__(self):
        return "<class: '" + str(self.__class__.__name__) + "'>: " + str(list(self))

class PacketHeader6(PrettyPrint):
    def __init__(
            self,
            flags: Optional[PacketFlags6] = None,
            ack: int = 0,
            token: bytes = b'\xff\xff\xff\xff',
            num_chunks: Optional[int] = None
    ) -> None:
        """
        If num_chunks is not set it will count
        the messages it was given when
        the pack() method is called
        """
        if not flags:
            flags = PacketFlags6()
        self.flags: PacketFlags6 = flags
        self.ack: int = ack % NET_MAX_SEQUENCE

        # the token has to be 4 bytes at all times
        # ff ff ff ff represents the empty token
        # if the packet is a 0.6.5 packet the token will be part of the header
        # if the packet is a 0.6.4 ddnet packet the token will be appended at the
        # end of the packet after the payload
        self.token: bytes = token
        self.num_chunks: Optional[int] = num_chunks

        # connless only
        self.connless_version: int = NET_PACKETVERSION
        self.response_token: bytes = b'\xff\xff\xff\xff'

    def __iter__(self):
        yield 'flags', list(self.flags)
        yield 'ack', self.ack
        yield 'token', self.token
        yield 'num_chunks', self.num_chunks

        if self.flags.connless:
            yield 'connless_version', self.connless_version
            yield 'response_token', self.response_token

    def pack(self, include_token: bool = True) -> bytes:
        """
        Generate 7 byte teeworlds 0.6.5 packet header
        based on the current instance variable
        values.

        The layout is as follows
        6bit flags, 2bit ack
        8bit ack
        8bit chunks
        32bit token

        ffffffaa
        aaaaaaaa
        NNNNNNNN
        TTTTTTTT
        TTTTTTTT
        TTTTTTTT
        TTTTTTTT
        """
        flags = 0
        if self.flags.token is None:
            # do not automatically set the token flag
            # if the token field has the empty token value
            if self.token != b'\xff\xff\xff\xff':
                self.flags.token = True
        if self.flags.token and include_token:
            flags |= PACKETFLAG6_TOKEN
        if self.flags.control:
            flags |= PACKETFLAG6_CONTROL
        if self.flags.connless:
            flags |= PACKETFLAG6_CONNLESS
        if self.flags.resend:
            flags |= PACKETFLAG6_RESEND
        if self.flags.compression:
            flags |= PACKETFLAG6_COMPRESSION
        if self.num_chunks is None:
            self.num_chunks = 0
        if self.flags.connless:
            return b'\xff\xff\xff\xff\xff\xff'
        packed = bytes([ \
            ((flags << 2)&0xfc) | ((self.ack>>8)&0x03), \
            self.ack&0xff, \
            self.num_chunks \
        ])
        if self.flags.token and include_token:
            packed += self.token
        return packed

class TwPacket6(PrettyPrint):
    def __init__(self, is_ddnet: bool = False, version: Literal['0.6']= '0.6') -> None:
        # teeworlds protocol version
        # can be set to a more specific version than 0.6 using set_version()
        self._version: Literal['0.6', '0.6.4', '0.6.5'] = version

        # does this packet use the 0.6.4 + TKEN extension protocol
        # if true the last 4 bytes of the packet are the ddnet security token
        self.is_ddnet = is_ddnet

        # packet payload as raw bytes
        # might be huffman compressed if header.flags.compression is set
        self.payload_raw: bytes = b''

        # packet payload after huffman decompression
        # can match payload_raw if no compression was applied
        self.payload_decompressed: bytes = b''

        # information about the packet
        # metadata that is not contained in the payload
        self.header = PacketHeader6()

        # packet payload parsed as message array
        self.messages: list[Union[CtrlMessage, NetMessage, ConnlessMessage]] = []

    def __iter__(self) -> Iterator[tuple[str, Any]]:
        yield 'version', self.version
        yield 'is_ddnet', self.is_ddnet
        yield 'payload_raw', self.payload_raw
        yield 'payload_decompressed', self.payload_decompressed
        yield 'header', dict(self.header)

        yield 'messages', [dict(msg) for msg in self.messages]

    def to_json(self) -> str:
        return json.dumps(
            dict(self),
            indent=2,
            sort_keys=False,
            default=twnet_parser.serialize.bytes_to_hex
        )

    @property
    def version(self) -> Literal['0.6', '0.6.4', '0.6.5']:
        has_token = self.header.token != b'\xff\xff\xff\xff' or self.header.flags.token
        if self.is_ddnet:
            self._version = '0.6.4'
        elif has_token:
            self._version = '0.6.5'
        return self._version

    def set_version(self, version: Literal['0.6', '0.6.4', '0.6.5']) -> str:
        self._version = version
        return self._version

    def pack(self, we_are_a_client = True) -> bytes:
        payload: bytes = b''
        msg: Union[CtrlMessage, NetMessage, ConnlessMessage]
        is_control: bool = False
        is_connless: bool = False
        for msg in self.messages:
            if msg.message_type == 'connless':
                is_connless = True
                msg = cast(ConnlessMessage, msg)
                payload += bytes(msg.message_id)
                payload += msg.pack()
            elif msg.message_type == 'control':
                is_control = True
                msg = cast(CtrlMessage, msg)
                payload += pack_int(msg.message_id)
                payload += msg.pack(we_are_a_client)
            else: # game or system message
                msg = cast(NetMessage, msg)
                msg_payload: bytes = pack_int(
                    (msg.message_id<<1) |
                    (int)(msg.system_message)
                )
                msg_payload += msg.pack()
                if msg.header.size is None:
                    msg.header.size = len(msg_payload)
                payload += msg.header.pack()
                payload += msg_payload
        if self.header.num_chunks is None:
            if is_control:
                self.header.num_chunks = 0
            else:
                self.header.num_chunks = len(self.messages)
        if is_control:
            if self.header.flags.control is None:
                self.header.flags.control = True
        if is_connless:
            if self.header.flags.connless is None:
                self.header.flags.connless = True
        if self.header.flags.compression:
            payload = twnet_parser.huffman.compress(payload)
        if self.is_ddnet:
            return self.header.pack(include_token=False) + payload + self.header.token
        return self.header.pack() + payload

class PacketHeaderParser6():
    def parse_flags6(self, data: bytes) -> PacketFlags6:
        # FFFF FFaa
        flag_bits = data[0] >> 2
        flags = PacketFlags6()
        flags.token = (flag_bits & PACKETFLAG6_TOKEN) != 0
        flags.control = (flag_bits & PACKETFLAG6_CONTROL) != 0
        flags.connless = (flag_bits & PACKETFLAG6_CONNLESS) != 0
        flags.resend = (flag_bits & PACKETFLAG6_RESEND) != 0
        flags.compression = (flag_bits & PACKETFLAG6_COMPRESSION) != 0
        if flags.connless:
            # connless packets send FF
            # as the flag byte
            # but setting the connless bit basically means
            # all other flags are false implicitly
            flags.token = False
            flags.control = False
            flags.resend = False
            flags.compression = False
        return flags

    def parse_ack(self, header_bytes: bytes) -> int:
        # ffAA AAAA AAAA
        return ((header_bytes[0] & 0x3) << 8) | header_bytes[1]

    def parse_num_chunks(self, header_bytes: bytes) -> int:
        # TODO: not sure if this is correct
        return header_bytes[2]

    def parse_token(self, header_bytes: bytes) -> bytes:
        return header_bytes[3:7]

    def parse_header(self, data: bytes) -> PacketHeader6:
        header = PacketHeader6()
        # bits 1..5
        header.flags = self.parse_flags6(data)
        if header.flags.connless:
            # TODO: do not hardcode version field
            #       but actually read the bits
            header.connless_version = NET_PACKETVERSION
            header.token = data[1:5]
            header.response_token = data[5:9]
        else:
            # bits 6..16
            header.ack = self.parse_ack(data)
            # bits 17..25
            header.num_chunks = self.parse_num_chunks(data)
            # bits 16..57
            if header.flags.token:
                header.token = self.parse_token(data)
            else:
                header.token = b''
        return header




# generated by scripts/generate_messages.py

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.chunk_header import ChunkHeader
from twnet_parser.packer import pack_int, pack_str
from typing import Literal, Optional
import twnet_parser

class MsgInfo(PrettyPrint):
    def __init__(
            self,
            chunk_header: Optional[ChunkHeader] = None,
            version: str = '0.7 802f1be60a05665f',
            password: str = '',
            client_version: int = 1797
    ) -> None:
        self.message_type: Literal['system', 'game'] = 'system'
        self.message_name: str = 'info'
        self.system_message: bool = True
        self.message_id: int = 1
        if not chunk_header:
            chunk_header = ChunkHeader(version = '0.7')
        self.header: ChunkHeader = chunk_header

        self.version: str = version
        self.password: Optional[str] = password
        self.client_version: Optional[int] = client_version

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'system_message', self.system_message
        yield 'message_id', self.message_id
        yield 'header', dict(self.header)

        yield 'version', self.version
        yield 'password', self.password
        yield 'client_version', self.client_version

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        self.version = unpacker.get_str()
        self.password = unpacker.get_optional_str()
        self.client_version = unpacker.get_optional_int()
        return True

    def pack(self) -> bytes:
        lib_version = b''
        if self.password is not None and self.client_version is not None:
            lib_version = pack_str(f"twnet_parser {twnet_parser.__version__}")
        return pack_str(self.version) + \
            (pack_str(self.password) if self.password is not None else b'') + \
            (pack_int(self.client_version) if self.client_version is not None else b'') + \
            lib_version



# generated by scripts/generate_messages.py

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.chunk_header import ChunkHeader
from twnet_parser.packer import SANITIZE_CC, pack_int, pack_str
from typing import Annotated, Literal, Optional
import twnet_parser.enum7 as enum7

class MsgSvClientInfo(PrettyPrint):
    def __init__(
            self,
            chunk_header: Optional[ChunkHeader] = None,
            client_id: int = 0,
            local: bool = False,
            team: int = enum7.Team.SPECTATORS.value,
            name: str = 'default',
            clan: str = 'default',
            country: int = 0,
            skin_part_names: Annotated[list[str], 6] = \
                ['', '', '', '', '', ''],
            use_custom_colors: Annotated[list[bool], 6] = \
                [False, False, False, False, False, False],
            skin_part_colors: Annotated[list[int], 6] = \
                [0, 0, 0, 0, 0, 0],
            silent: bool = False
    ) -> None:
        self.message_type: Literal['system', 'game'] = 'game'
        self.message_name: str = 'sv_client_info'
        self.system_message: bool = False
        self.message_id: int = 18
        if not chunk_header:
            chunk_header = ChunkHeader(version = '0.7')
        self.header: ChunkHeader = chunk_header

        self.client_id: int = client_id
        self.local: bool = local
        self.team: int = team
        self.name: str = name
        self.clan: str = clan
        self.country: int = country
        self.skin_part_names = skin_part_names
        self.use_custom_colors = use_custom_colors
        self.skin_part_colors = skin_part_colors
        self.silent: bool = silent

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'system_message', self.system_message
        yield 'message_id', self.message_id
        yield 'header', dict(self.header)

        yield 'client_id', self.client_id
        yield 'local', self.local
        yield 'team', self.team
        yield 'name', self.name
        yield 'clan', self.clan
        yield 'country', self.country
        yield 'skin_part_names', self.skin_part_names
        yield 'use_custom_colors', self.use_custom_colors
        yield 'skin_part_colors', self.skin_part_colors
        yield 'silent', self.silent

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        self.client_id = unpacker.get_int()
        self.local = unpacker.get_int() == 1
        self.team = unpacker.get_int() # enum TEAM
        self.name = unpacker.get_str(SANITIZE_CC)
        self.clan = unpacker.get_str(SANITIZE_CC)
        self.country = unpacker.get_int()
        for i in range(0, 6):
            self.skin_part_names[i] = unpacker.get_str(SANITIZE_CC)
        for i in range(0, 6):
            self.use_custom_colors[i] = unpacker.get_int() == 1
        for i in range(0, 6):
            self.skin_part_colors[i] = unpacker.get_int()
        self.silent = unpacker.get_int() == 1
        return True

    def pack(self) -> bytes:
        return pack_int(self.client_id) + \
            pack_int(self.local) + \
            pack_int(self.team) + \
            pack_str(self.name) + \
            pack_str(self.clan) + \
            pack_int(self.country) + \
            b''.join([pack_str(x) for x in self.skin_part_names]) + \
            b''.join([pack_int(x) for x in self.use_custom_colors]) + \
            b''.join([pack_int(x) for x in self.skin_part_colors]) + \
            pack_int(self.silent)


# generated by scripts/generate_messages.py

from typing import Any, Dict

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import pack_int
from typing import Annotated

class ObjDeTuneParams(PrettyPrint):
    def __init__(
            self,
            tune_params: Annotated[list[int], 32] = \
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    ) -> None:
        self.item_name: str = 'obj.de_tune_params'
        self.type_id: int = 15
        self.id: int = 0
        self.size: int = 1

        self.tune_params = tune_params

    def __iter__(self):
        yield 'item_name', self.item_name
        yield 'type_id', self.type_id
        yield 'id', self.id
        yield 'size', self.size

        yield 'tune_params', self.tune_params

    def to_dict_payload_only(self) -> Dict[str, Any]:
        return {
            'tune_params': self.tune_params
        }

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, unpacker: Unpacker) -> bool:
        for i in range(0, 32):
            self.tune_params[i] = unpacker.get_int()
        return True

    def pack(self) -> bytes:
        return b''.join([pack_int(x) for x in self.tune_params])